"""Build tools for bertrand-enabled C++ extensions."""
import re
import shutil
import subprocess
import sys
import sysconfig
from pathlib import Path
from typing import Any

import numpy
import pybind11
from pybind11.setup_helpers import Pybind11Extension
from pybind11.setup_helpers import build_ext as pybind11_build_ext
import setuptools


# TODO: list cpptrace, pcre2, googletest version #s explicitly here?

ROOT: Path = Path(__file__).absolute().parent.parent
DEPS: Path = ROOT / "third_party"


def get_include() -> str:
    """Get the path to the include directory for this package, which is necessary to
    make C++ headers available to the compiler.

    Returns
    -------
    str
        The path to the include directory for this package.
    """
    return str(ROOT)


def quick_include() -> list[str]:
    """Return the complete include and link flags necessary to build a pure-C++ project
    with bertrand as a dependency.

    Returns
    -------
    list
        A list of strings containing the various include and link libraries needed to
        build a bertrand-enabled project from the command line, as a single unit.
    """
    cpptrace = DEPS / "cpptrace-0.5.2"
    gtest = DEPS / "googletest-1.14.0"
    pcre2 = DEPS / "pcre2-10.43"
    return [
        f"-I{sysconfig.get_path('include')}",
        f"-I{get_include()}",
        f"-I{pybind11.get_include()}",
        f"-I{numpy.get_include()}",
        f"-I{str(cpptrace / 'include')}",
        f"-I{str(gtest / 'googletest' / 'include')}",
        f"-I{str(pcre2 / 'src')}",
        f"-L{str(cpptrace / 'build')}",
        f"-L{str(cpptrace / 'build' / '_deps' / 'libdwarf-build' / 'src' / 'lib' / 'libdwarf')}",
        f"-L{str(cpptrace / 'build' / '_deps' / 'zstd-build' / 'lib')}",
        f"-L{str(gtest / 'build' / 'lib')}",
        f"-L{str(pcre2 / '.libs')}",
        f"-lpython{sysconfig.get_python_version()}",
        "-lcpptrace",
        "-ldwarf",
        "-lz",
        "-lzstd",
        "-ldl",
        "-lgtest",
    ]


class Extension(Pybind11Extension):
    """A setuptools.Extension class that builds using CMake and supports C++20 modules.

    Parameters
    ----------
    *args, **kwargs : Any
        Arbitrary arguments passed to the Pybind11Extension constructor.
    cxx_std : int, default 20
        The C++ standard to use when compiling the extension.  Values less than 20 will
        raise a ValueError.
    traceback : bool, default True
        If set to false, add `BERTRAND_NO_TRACEBACK` to the compile definitions, which
        will disable cross-language tracebacks for the extension.
    cmake_args : dict[str, Any], optional
        Additional arguments to pass to the Extension's CMake configuration.  These are
        emitted as key-value pairs into a `set_target_properties()` block in the
        generated CMakeLists.txt file.  Some options are filled in by default,
        including `PREFIX`, `LIBRARY_OUTPUT_DIRECTORY`, `LIBRARY_OUTPUT_NAME`,
        `SUFFIX`, `CXX_STANDARD`, and `CXX_STANDARD_REQUIRED`.
    """

    MODULE_REGEX = re.compile(r"\s*export\s+module\s+(\w+).*;", re.MULTILINE)

    def __init__(
        self,
        *args: Any,
        executable: bool = False,
        cxx_std: int = 23,
        traceback: bool = True,
        cmake_args: dict[str, Any] | None = None,
        **kwargs: Any
    ) -> None:
        if cxx_std < 23:
            raise ValueError(
                "C++ standard must be at least C++23 to enable bertrand features"
            )

        super().__init__(*args, **kwargs)
        self.executable = executable
        self.cxx_std = cxx_std
        self.traceback = traceback
        self.cmake_args = cmake_args or {}

        self.include_dirs.append(get_include())
        self.include_dirs.append(numpy.get_include())
        if self.traceback:
            self.extra_compile_args.append("-g")
            self.extra_link_args.append("-g")
        else:
            self.define_macros.append("BERTRAND_NO_TRACEBACK")

    def add_to_cmakelists(
        self,
        cmakelists: Path,
        build_dir: Path,
        module_cache: dict[str, bool]
    ) -> None:
        """Generate a temporary CMakeLists.txt that configures the extension for
        building with CMake.

        Parameters
        ----------
        cmakelists : Path
            The (current) path to the generated CMakeLists.txt file.
        build_dir : Path
            The path to the build directory for the extension.
        module_cache : dict[str, bool]
            A dictionary that caches whether each source file exports a module.  If the
            same source is used across multiple extensions, this will prevent duplicate
            scans.
        """
        with cmakelists.open("a") as f:
            if self.executable:
                f.write(f"add_executable({self.name}\n")
            else:
                f.write(f"add_library({self.name} MODULE\n")
            # TODO: add sources related to built-in bertrand modules
            for source in self.sources:
                f.write(f"    {ROOT / source}\n")
            f.write(")\n")

            f.write(f"set_target_properties({self.name} PROPERTIES\n")
            f.write( "    PREFIX \"\"\n")
            f.write(f"    OUTPUT_NAME {self.name}\n")
            if not self.executable:
                f.write(f"    LIBRARY_OUTPUT_DIRECTORY {build_dir}")
                f.write(f"    SUFFIX {sysconfig.get_config_var('EXT_SUFFIX')}\n")
            f.write(f"    CXX_STANDARD {self.cxx_std}\n")
            f.write( "    CXX_STANDARD_REQUIRED ON\n")
            for key, value in self.cmake_args.items():
                f.write(f"    {key} {value}\n")
            f.write(")\n")

            modules = []
            for source in self.sources:
                if source not in module_cache:
                    with Path(ROOT / source).open("r", encoding="utf_8") as s:
                        module_cache[source] = bool(self.MODULE_REGEX.search(s.read()))
                if module_cache[source]:
                    modules.append(source)

            if modules:
                f.write(f"target_sources({self.name} PRIVATE\n")
                f.write( "    FILE_SET CXX_MODULES\n")
                f.write(f"    BASE_DIRS {ROOT}\n")
                f.write( "    FILES\n")
                for source in modules:
                    f.write(f"        {ROOT / source}\n")
                f.write(")\n")

            if self.include_dirs:
                f.write(f"target_include_directories({self.name} PRIVATE\n")
                for include in self.include_dirs:
                    f.write(f"    {include}\n")
                f.write(")\n")

            if self.library_dirs:
                f.write(f"target_link_directories({self.name} PRIVATE\n")
                for lib_dir in self.library_dirs:
                    f.write(f"    {lib_dir}\n")
                f.write(")\n")

            if self.libraries:
                f.write(f"target_link_libraries({self.name} PRIVATE\n")
                for lib in self.libraries:
                    f.write(f"    {lib}\n")
                f.write(")\n")

            if self.extra_compile_args:
                f.write(f"target_compile_options({self.name} PRIVATE\n")
                for flag in self.extra_compile_args:
                    f.write(f"    {flag}\n")
                f.write(")\n")

            _link_options = [sysconfig.get_config_var("LDFLAGS")] + self.extra_link_args
            if _link_options:
                f.write(f"target_link_options({self.name} PRIVATE\n")
                for flag in _link_options:
                    f.write(f"    {flag}\n")
                f.write(")\n")

            if self.define_macros:
                f.write(f"target_compile_definitions({self.name} PRIVATE\n")
                for define in self.define_macros:
                    if isinstance(define, tuple):
                        f.write(f"    {define[0]}={define[1]}\n")
                    else:
                        f.write(f"    {define}\n")
                f.write(")\n")


class BuildExt(pybind11_build_ext):
    """A custom build_ext command that uses CMake to build extensions with support for
    C++20 modules, parallel builds, clangd, executable targets, and bertrand's core
    dependencies without any extra configuration.
    """

    CMAKE_MIN_VERSION = "3.28" # CMake 3.28+ is necessary for C++20 module support
    user_options = pybind11_build_ext.user_options + [
        (
            "workers=",
            "j",
            "The number of parallel workers to use when building CMake extensions"
        )
    ]

    def __init__(self, *args: Any, workers: int | None = None, **kwargs: Any) -> None:
        """Initialize the build extensions command.

        Parameters
        ----------
        *args : Any
            Arbitrary positional arguments passed to the parent class.
        workers : int, default 0
            The number of parallel workers to use when building the extensions.  If set
            to 0, then the build will be single-threaded.
        **kwargs : Any
            Arbitrary keyword arguments passed to the parent class.
        """
        super().__init__(*args, **kwargs)
        self.workers = workers

    def finalize_options(self) -> None:
        """Extract the number of parallel workers from the setup() call.

        Raises
        ------
        ValueError
            If the workers option is not set to a positive integer.
        """
        super().finalize_options()
        if self.workers is not None:
            try:
                self.workers = int(self.workers)
            except ValueError as e:
                raise ValueError("workers must be set to an integer") from e

            if self.workers < 1:
                raise ValueError("workers must be set to a positive integer")

    def init_cmakelists(self) -> Path:
        """Create and initialize a CMakeLists.txt file for the entire project.

        Returns
        -------
        Path
            The path to the generated CMakeLists.txt file, for further processing.
        """
        file = Path(self.build_lib).absolute() / "CMakeLists.txt"
        file.parent.mkdir(parents=True, exist_ok=True)
        file.touch(exist_ok=True)

        with file.open("w") as f:
            f.write(f"cmake_minimum_required(VERSION {self.CMAKE_MIN_VERSION})\n")
            f.write(f"project({self.distribution.get_name()} LANGUAGES CXX)\n")
            f.write(f"set(CMAKE_BUILD_TYPE {'Debug' if self.debug else 'Release'})\n")
            f.write(f"set(PYTHON_EXECUTABLE {sys.executable})\n")
            f.write("set(CMAKE_COLOR_DIAGNOSTICS ON)\n")
            f.write("set(CMAKE_CXX_SCAN_FOR_MODULES ON)\n")
            f.write("set(CMAKE_EXPORT_COMPILE_COMMANDS ON)\n")

            if self.compiler.include_dirs:
                f.write("include_directories(\n")
                for include in self.compiler.include_dirs:
                    f.write(f"    {include}\n")
                f.write(")\n")

            if self.compiler.library_dirs:
                f.write("link_directories(\n")
                for lib_dir in self.compiler.library_dirs:
                    f.write(f"    {lib_dir}\n")
                f.write(")\n")

            if self.compiler.libraries:
                f.write("link_libraries(\n")
                f.write(f"    python{sysconfig.get_python_version()}\n")
                for lib in self.compiler.libraries:
                    f.write(f"    {lib}\n")
                f.write(")\n")

        return file

    def build_extensions(self) -> None:
        """Build all extensions in the project."""
        self.check_extensions_list(self.extensions)

        call_cmake = False
        cmakelists = self.init_cmakelists()
        module_cache: dict[str, bool] = {}
        for ext in self.extensions:
            if isinstance(ext, Extension):
                call_cmake = True
                build_dir = Path(ROOT / self.get_ext_fullpath(ext.name)).parent
                ext.add_to_cmakelists(cmakelists, build_dir, module_cache)
            else:
                super().build_extension(ext)

        config_args = [
            "cmake",
            "-G",
            "Ninja",
            str(cmakelists.parent)
        ]
        build_args = [
            "cmake",
            "--build",
            ".",
            "--config",
            "Debug" if self.debug else "Release"
        ]
        if self.workers:
            build_args += ["--parallel", str(self.workers)]

        print(self.workers)

        if call_cmake:
            try:
                subprocess.check_call(config_args, cwd=cmakelists.parent)
                subprocess.check_call(build_args, cwd=cmakelists.parent)
            except subprocess.CalledProcessError as e:
                if e.stderr:
                    print(e.stderr)

            shutil.move(
                cmakelists.parent / "compile_commands.json",
                ROOT / "compile_commands.json"
            )

    def get_ext_filename(self, fullname: str) -> str:
        """Format the filename of the extension module.

        Parameters
        ----------
        fullname : str
            The name of the extension module.

        Returns
        -------
        str
            The formatted filename of the build product (either a shared library or an
            executable).  If the extension is an executable, the name is returned
            directly.  Otherwise, it will be decorated with a shared library extension
            based on the platform, like normal.

        Notes
        -----
        Overriding this method is required to allow setuptools to correctly copy the
        build products out of the build directory without errors.
        """
        ext = self.ext_map[fullname]
        if getattr(ext, "executable", False):
            return ext.name
        return super().get_ext_filename(fullname)


def setup(
    *args: Any,
    cmdclass: dict[str, Any] | None = None,
    workers: int | None = None,
    **kwargs: Any
) -> None:
    """A custom setup() function that automatically appends the BuildExt command to the
    setup commands.

    Parameters
    ----------
    *args : Any
        Arbitrary positional arguments passed to the setuptools.setup() function.
    cmdclass : dict[str, Any] | None, default None
        A dictionary of command classes to override the default setuptools commands.
        If no setting is given for "build_ext", then it will be set to
        bertrand.setuptools.BuildExt.
    workers : int | None, default None
        The number of parallel workers to use when building extensions.  If set to
        None, then the build will be single-threaded.  This can also be set through
        the command line by supplying either `--workers=NUM` or `-j NUM`.
    **kwargs : Any
        Arbitrary keyword arguments passed to the setuptools.setup() function.
    """
    class BuildExtensions(BuildExt):
        """A private subclass of BuildExt that captures the number of workers to use
        from the setup() call.
        """
        def __init__(self, *a: Any, **kw: Any):
            super().__init__(*a, workers=workers, **kw)

    if cmdclass is None:
        cmdclass = {"build_ext": BuildExtensions}
    elif "build_ext" not in cmdclass:
        cmdclass["build_ext"] = BuildExtensions

    setuptools.setup(*args, cmdclass=cmdclass, **kwargs)



"""This module contains dispatched cast() implementations for integer data."""
# pylint: disable=unused-argument
from __future__ import annotations
import datetime
from functools import partial

import numpy as np
import pandas as pd

from pdcast import types
from pdcast.detect import detect_type
from pdcast.resolve import resolve_type
from pdcast.util import time
from pdcast.util.numeric import boundscheck
from pdcast.util.round import round_div, Tolerance
from pdcast.util.string import int_to_base
from pdcast.util.vector import apply_with_errors

from .base import (
    cast, generic_to_boolean, generic_to_integer, generic_to_float,
    generic_to_string
)


# TODO: int -> timedelta is unusually slow for some reason


@cast.overload("int", "bool")
def integer_to_boolean(
    series: pd.Series,
    dtype: types.VectorType,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a boolean data type."""
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)
    return generic_to_boolean(series, dtype, errors=errors)


@cast.overload("int", "int")
def integer_to_integer(
    series: pd.Series,
    dtype: types.VectorType,
    tol: Tolerance,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to another integer data type."""
    # trivial case
    if detect_type(series) == dtype:
        return series

    # check for overflow and upcast if applicable
    try:
        series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)
    except OverflowError as err:
        last_err = err
        for typ in dtype.larger:
            try:
                series = boundscheck(
                    series,
                    dtype=typ,
                    tol=tol.real,
                    errors=errors
                )
                dtype = typ
                break
            except OverflowError as exc:
                last_err = exc
        else:
            raise last_err from err

    # use generic implementation
    return generic_to_integer(
        series,
        dtype,
        tol=tol,
        downcast=downcast,
        errors=errors
    )


@cast.overload("int", "float")
def integer_to_float(
    series: pd.Series,
    dtype: types.VectorType,
    tol: Tolerance,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a float data type."""
    # NOTE: integers can always be exactly represented as floats as long as
    # their width in bits fits within the significand of the specified floating
    # point type with exponent 1 (as listed in the IEEE 754 specification).

    if int(series.min()) < dtype.min or int(series.max()) > dtype.max:
        # 2-step conversion: int -> decimal, decimal -> float
        series = cast(series, "decimal", errors=errors)
        return cast(
            series,
            dtype,
            tol=tol,
            downcast=downcast,
            errors=errors,
            **unused
        )

    return generic_to_float(
        series,
        dtype,
        tol=tol,
        downcast=downcast,
        errors=errors
    )


@cast.overload("int", "complex")
def integer_to_complex(
    series: pd.Series,
    dtype: types.VectorType,
    tol: Tolerance,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a complex data type."""
    # 2-step conversion: int -> float, float -> complex
    series = cast(
        series,
        "float",
        tol=tol,
        downcast=None,
        errors=errors
    )
    return cast(
        series,
        dtype,
        tol=tol,
        downcast=downcast,
        errors=errors,
        **unused
    )


@cast.overload("int", "decimal")
def integer_to_decimal(
    series: pd.Series,
    dtype: types.VectorType,
    **unused
) -> pd.Series:
    """Convert integer data to a decimal data type."""
    target = dtype.dtype
    if isinstance(target, types.ObjectDtype):
        series = series + dtype.type_def(0)  # ~2x faster than apply loop
    return series.astype(target)


@cast.overload("int", "datetime")
def integer_to_datetime(
    series: pd.Series,
    dtype: types.VectorType,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tz: datetime.tzinfo,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a datetime data type."""
    # convert to ns
    series = to_ns(series, unit=unit, step_size=step_size, since=since)

    # account for non-utc epoch
    if since:
        series += since.offset

    # check for overflow and upcast if applicable
    series = boundscheck(series, dtype, tol=tol.real, errors=errors)

    # convert to final representation
    return cast(
        series,
        dtype.registry.get_default(dtype),
        unit="ns",
        step_size=1,
        since=time.Epoch("utc"),
        tz=tz,
        errors=errors,
        **unused
    )


@cast.overload("int", "datetime[pandas]")
def integer_to_pandas_timestamp(
    series: pd.Series,
    dtype: types.VectorType,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tz: datetime.tzinfo,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a datetime data type."""
    # convert to ns
    series = to_ns(series, unit=unit, step_size=step_size, since=since)

    # account for non-utc epoch
    if since:
        series += since.offset

    # check for overflow and upcast if applicable
    series = boundscheck(series, dtype, tol=tol.real, errors=errors)

    # reconcile `tz` argument with timezone attached to dtype, if given
    if tz:
        dtype = dtype.replace(tz=tz)

    # convert using pd.to_datetime, accounting for timezone
    if dtype.tz is None:
        series = pd.to_datetime(series, unit="ns")
    else:
        series = pd.to_datetime(series, unit="ns", utc=True)
        if not time.is_utc(dtype.tz):
            series = series.dt.tz_convert(dtype.tz)

    return series


@cast.overload("int", "datetime[python]")
def integer_to_python_datetime(
    series: pd.Series,
    dtype: types.VectorType,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tz: datetime.tzinfo,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a datetime data type."""
    # convert to ns
    series = to_ns(series, unit=unit, step_size=step_size, since=since)

    # account for non-utc epoch
    if since:
        series += since.offset

    # check for overflow and upcast if applicable
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # reconcile `tz` argument with timezone attached to dtype, if given
    if tz:
        dtype = dtype.replace(tz=tz)

    # convert elementwise
    call = partial(time.ns_to_pydatetime, tz=dtype.tz)
    result = apply_with_errors(series, call, errors=errors)
    return result.astype(dtype.dtype)


@cast.overload("int", "datetime[numpy]")
def integer_to_numpy_datetime64(
    series: pd.Series,
    dtype: types.VectorType,
    unit: str,
    step_size: int,
    rounding: str,
    since: time.Epoch,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a datetime data type."""
    # convert to ns
    series = to_ns(series, unit=unit, step_size=step_size, since=since)

    # account for non-utc epoch
    if since:
        series += since.offset

    # check for overflow and upcast if applicable
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # convert from nanoseconds to final representation
    series = time.convert_unit(
        series,
        "ns",
        dtype.unit,
        rounding=rounding or "down"
    )
    if dtype.step_size != 1:
        series = round_div(
            series,
            dtype.step_size,
            rule=rounding or "down"
        )

    M8_str = f"M8[{dtype.step_size}{dtype.unit}]"
    return pd.Series(
        list(series.to_numpy(M8_str)),
        index=series.index,
        dtype=object
    )


@cast.overload("int", "timedelta")
def integer_to_timedelta(
    series: pd.Series,
    dtype: types.VectorType,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a timedelta data type."""
    # convert to ns
    series = to_ns(series, unit=unit, step_size=step_size, since=since)

    # check for overflow and upcast if necessary
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # convert to final representation
    return cast(
        series,
        dtype.registry.get_default(dtype),
        unit="ns",
        step_size=1,
        since=since,
        errors=errors,
        **unused
    )


@cast.overload("int", "timedelta[pandas]")
def integer_to_pandas_timedelta(
    series: pd.Series,
    dtype: types.VectorType,
    unit: str,
    step_size: int,
    since: time.Epoch,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a timedelta data type."""
    series = to_ns(series, unit=unit, step_size=step_size, since=since)

    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # NOTE: pandas.to_timedelta complains when given an ObjectArray as input,
    # so we convert to dtype: object instead.

    return pd.to_timedelta(series.astype(object), unit="ns")


@cast.overload("int", "timedelta[python]")
def integer_to_python_timedelta(
    series: pd.Series,
    dtype: types.VectorType,
    unit: str,
    step_size: int,
    rounding: str,
    since: time.Epoch,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a timedelta data type."""
    # convert to ns
    series = to_ns(series, unit=unit, step_size=step_size, since=since)

    # check for overflow and upcast if necessary
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # convert to us
    result = round_div(series, time.as_ns["us"], rule=rounding or "down")

    # NOTE: m8[us].astype(object) implicitly converts to datetime.timedelta

    return pd.Series(
        result.to_numpy("m8[us]").astype(object),
        index=series.index,
        dtype=dtype.dtype
    )


@cast.overload("int", "timedelta[numpy]")
def integer_to_numpy_timedelta64(
    series: pd.Series,
    dtype: types.VectorType,
    unit: str,
    step_size: int,
    rounding: str,
    since: time.Epoch,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a timedelta data type."""
    # convert to ns
    series = to_ns(series, unit=unit, step_size=step_size, since=since)

    # check for overflow and upcast if necessary
    series = boundscheck(series, dtype=dtype, tol=tol.real, errors=errors)

    # convert from ns to final unit
    series = time.convert_unit(
        series,
        "ns",
        dtype.unit,
        rounding=rounding or "down",
        since=since
    )
    if dtype.step_size != 1:
        series = round_div(
            series,
            dtype.step_size,
            rule=rounding or "down"
        )
    m8_str = f"m8[{dtype.step_size}{dtype.unit}]"
    return pd.Series(
        list(series.to_numpy(m8_str)),
        index=series.index,
        dtype="O"
    )


@cast.overload("int", "string")
def integer_to_string(
    series: pd.Series,
    dtype: types.VectorType,
    base: int,
    format: str,
    errors: str,
    **unused
) -> pd.Series:
    """Convert integer data to a string data type in any base."""
    # use non-decimal base in conjunction with format
    if base and base != 10:
        if format:
            call = lambda x: f"{int_to_base(x, base=base):{format}}"
        else:
            call = partial(int_to_base, base=base)

        result = apply_with_errors(series, call, errors=errors)
        return result.astype(dtype.dtype)

    return generic_to_string(
        series,
        dtype,
        format=format,
        errors=errors
    )


#######################
####    PRIVATE    ####
#######################


as_pyint = np.frompyfunc(int, 1, 1)


def to_ns(
    series: pd.Series,
    unit: str,
    step_size: int,
    since: time.Epoch
) -> pd.Series:
    """Convert an integer number of time units into nanoseconds from a given
    epoch.
    """
    series = as_pyint(series).astype(resolve_type(int).dtype)  # overflow-safe

    # trivial case
    if unit == "ns" and step_size == 1:
        return series

    if step_size != 1:
        series *= step_size
    return time.convert_unit(series, unit, "ns", since=since)



"""This module provides PEP 484-style type hints for ``pdcast`` constructs.
"""
import datetime
import numbers
from typing import runtime_checkable, List, Protocol, Tuple, Union

import numpy as np
import numpy.typing
import pandas as pd


#######################
####    SCALARS    ####
#######################


numeric = Union[
    numbers.Number,
    np.number
]


datetime_like = Union[
    pd.Timestamp,
    datetime.datetime,
    np.datetime64
]


timedelta_like = Union[
    pd.Timedelta,
    datetime.timedelta,
    np.timedelta64
]


type_specifier = Union[
    type,
    str,
    np.dtype,
    pd.api.extensions.ExtensionDtype
]


dtype_like = Union[
    np.dtype,
    pd.api.extensions.ExtensionDtype
]


#########################
####    ITERABLES    ####
#########################


array_like = numpy.typing.ArrayLike


list_like = Union[
    List,
    Tuple,
    array_like
]


#########################
####    PROTOCOLS    ####
#########################


@runtime_checkable
class Descriptor(Protocol):
    def __get__(self, instance, owner): ...



from .epoch cimport Epoch


# constants
cdef dict as_ns
cdef tuple valid_units


# functions
cpdef object convert_unit(
    object val,
    str from_unit,
    str to_unit,
    str rounding = *,
    Epoch since = *
)
cpdef object round_years_to_ns(object years, Epoch since)
cpdef object round_months_to_ns(object months, Epoch since)
cdef object _convert_unit_irregular_to_regular(
    object val,
    str from_unit,
    str to_unit,
    str rounding,
    Epoch since
)
cdef object _convert_unit_regular_to_irregular(
    object val,
    str from_unit,
    str to_unit,
    str rounding,
    Epoch since
)



# """Flexible type extensions for pandas.

# Subpackages
# -----------
# convert
#     Extendable conversions between types in the ``pdcast`` type system.

# patch
#     Direct ``pdcast`` integration and type-aware attribute dispatch for
#     ``pandas.Series`` and ``pandas.DataFrame`` objects.

# types
#     Defines the structure and contents of the ``pdcast`` type system.

# util
#     Utilities for ``pdcast``-related functionality.

# Modules
# -------
# check
#     Fast type checks within the ``pdcast`` type system.

# detect
#     Type inference for arbitrary, vectorized data.

# resolve
#     Easy construction of data types from type specifiers, including a
#     domain-specific mini-language for referring to types.
# """
# # pylint: disable=undefined-variable, redefined-builtin
# from .convert import (
#     cast, categorize, decategorize, densify, sparsify, to_boolean, to_integer,
#     to_float, to_complex, to_decimal, to_datetime, to_timedelta, to_string
# )
# from .decorators.base import FunctionDecorator, Signature, Arguments
# from .decorators.attachable import (
#     attachable, Attachable, ClassMethod, InstanceMethod, Namespace, Property,
#     StaticMethod, VirtualAttribute
# )
# from .decorators.extension import (
#     extension_func, ExtensionFunc, ExtensionSignature, ExtensionArguments
# )
# from .decorators.dispatch import (
#     dispatch, DispatchFunc, DispatchSignature, DispatchArguments
# )
# from .patch.base import attach, detach


# # from .decorators import attach, dispatch, introspect
# # from .convert import cast
# from .types import *
# from .structs import LinkedList, LinkedSet, LinkedDict


# # importing * from types also masks module names, which can be troublesome
# del base            # type: ignore
# del boolean         # type: ignore
# del categorical     # type: ignore
# del complex
# # del datetime      # type: ignore
# del decimal         # type: ignore
# del float
# del integer         # type: ignore
# del missing         # type: ignore
# del object
# del sparse          # type: ignore
# del string          # type: ignore
# del timedelta       # type: ignore


from .setuptools import BuildExt, Extension, get_include, setup


try:
    from .regex import Regex
except ModuleNotFoundError:
    pass


__version__ = "0.5.1"  # handled by bumpver



"""This module describes a pair of functions that can be used to convert
to and from sparse data representations.
"""
from typing import Any

import pandas as pd

from pdcast.decorators.dispatch import dispatch
from pdcast.detect import detect_type


@dispatch("series")
def sparsify(series: pd.Series, fill_value: Any = None) -> pd.Series:
    """Transform a dense series into a sparse series with the given fill value.

    Parameters
    ----------
    series : pandas.Series
        The series to transform.
    fill_value : Any, optional
        The value to hold out of the sparse series.  If not given, the
        :attr:`na_value <pdcast.ScalarType.na_value>` of the inferred series
        type will be used instead.

    Returns
    -------
    pandas.Series
        A sparse series with the same values as the input.

    Notes
    -----
    This function is automatically invoked whenever :func:`cast() <pdcast.cast>`
    is called with a sparse type as a target.  The result will be computed
    at the scalar level and then made sparse using this function.

    Examples
    --------
    .. doctest::

        >>> import pandas as pd

        >>> sparsify(pd.Series([1, 2, 3]), fill_value=1)
        0    1
        1    2
        2    3
        dtype: Sparse[int64, <NA>]
        >>> _.sparse.density
        0.6666666666666666
    """
    # if no explicit fill value is given, use the series type's na_value
    if fill_value is None:
        fill_value = detect_type(series).na_value

    # astype() to a sparse dtype
    return series.astype(pd.SparseDtype(series.dtype, fill_value))


@dispatch("series")
def densify(series: pd.Series) -> pd.Series:
    """Transform a sparse series into a dense series.

    Parameters
    ----------
    series : pandas.Series
        The series to transform.

    Returns
    -------
    pandas.Series
        A dense series with the same values as the input.

    Notes
    -----
    This function is automatically invoked whenever :func:`cast() <pdcast.cast>`
    is called on a sparse series.  It is used to densify the input before
    converting to the target type.

    Examples
    --------
    .. doctest::

        >>> import pandas as pd

        >>> densify(pd.Series([1, 2, 3]).astype("Sparse[int]"))
        0    1
        1    2
        2    3
        dtype: int64
    """
    return series.sparse.to_dense()



"""This module contains all the prepackaged integer types for the ``pdcast``
type system.
"""
import numpy as np
import pandas as pd

from .base import REGISTRY, Type


class Int(Type):
    """Abstract integer type."""

    aliases = {"int", "integer"}


######################
####    SIGNED    ####
######################


@Int.default
class Signed(Int):
    """Abstract signed integer type."""

    aliases = {"signed", "signed int", "signed integer", "i"}


class PythonInt(Signed, backend="python"):
    """Python signed integer type."""

    aliases = {int}
    scalar = int
    max = np.inf
    min = -np.inf
    is_nullable = True
    missing = None


@Signed.default
class Int64(Signed):
    """Abstract 64-bit integer type."""

    aliases = {"int64", "i8"}
    max = 2**63 - 1
    min = -2**63


@Int64.default
class NumpyInt64(Int64, backend="numpy"):
    """Numpy 64-bit integer type."""

    aliases = {np.int64, np.dtype(np.int64)}
    dtype = np.dtype(np.int64)
    is_nullable = False


@NumpyInt64.nullable
class PandasInt64(Int64, backend="pandas"):
    """Pandas 64-bit integer type."""

    aliases = {pd.Int64Dtype}
    dtype = pd.Int64Dtype()
    is_nullable = True


class Int32(Signed):
    """Abstract 32-bit integer type."""

    aliases = {"int32", "i4"}
    max = 2**31 - 1
    min = -2**31


@Int32.default
class NumpyInt32(Int32, backend="numpy"):
    """Numpy 32-bit integer type."""

    aliases = {np.int32, np.dtype(np.int32)}
    dtype = np.dtype(np.int32)
    is_nullable = False


@NumpyInt32.nullable
class PandasInt32(Int32, backend="pandas"):
    """Pandas 32-bit integer type."""

    aliases = {pd.Int32Dtype}
    dtype = pd.Int32Dtype()
    is_nullable = True


class Int16(Signed):
    """Abstract 16-bit integer type."""

    aliases = {"int16", "i2"}
    max = 2**15 - 1
    min = -2**15


@Int16.default
class NumpyInt16(Int16, backend="numpy"):
    """Numpy 16-bit integer type."""

    aliases = {np.int16, np.dtype(np.int16)}
    dtype = np.dtype(np.int16)
    is_nullable = False


@NumpyInt16.nullable
class PandasInt16(Int16, backend="pandas"):
    """Pandas 16-bit integer type."""

    aliases = {pd.Int16Dtype}
    dtype = pd.Int16Dtype()
    is_nullable = True


class Int8(Signed):
    """Abstract 8-bit integer type."""

    aliases = {"int8", "i1"}
    max = 2**7 - 1
    min = -2**7


@Int8.default
class NumpyInt8(Int8, backend="numpy"):
    """Numpy 8-bit integer type."""

    aliases = {np.int8, np.dtype(np.int8)}
    dtype = np.dtype(np.int8)
    is_nullable = False



@NumpyInt8.nullable
class PandasInt8(Int8, backend="pandas"):
    """Pandas 8-bit integer type."""

    aliases = {pd.Int8Dtype}
    dtype = pd.Int8Dtype()
    is_nullable = True


########################
####    UNSIGNED    ####
########################


class Unsigned(Int):
    """Abstract unsigned integer type."""

    aliases = {"unsigned", "unsigned int", "unsigned integer", "u"}


@Unsigned.default
class UInt64(Unsigned):
    """Abstract 64-bit unsigned integer type."""

    aliases = {"uint64", "unsigned int64", "u8"}
    max = 2**64 - 1
    min = 0


@UInt64.default
class NumpyUInt64(UInt64, backend="numpy"):
    """Numpy 64-bit unsigned integer type."""

    aliases = {np.uint64, np.dtype(np.uint64)}
    dtype = np.dtype(np.uint64)
    is_nullable = False


@NumpyUInt64.nullable
class PandasUInt64(UInt64, backend="pandas"):
    """Pandas 64-bit unsigned integer type."""

    aliases = {pd.UInt64Dtype}
    dtype = pd.UInt64Dtype()
    is_nullable = True


class UInt32(Unsigned):
    """Abstract 32-bit unsigned integer type."""

    aliases = {"uint32", "unsigned int32", "u4"}
    max = 2**32 - 1
    min = 0


@UInt32.default
class NumpyUInt32(UInt32, backend="numpy"):
    """Numpy 32-bit unsigned integer type."""

    aliases = {np.uint32, np.dtype(np.uint32)}
    dtype = np.dtype(np.uint32)
    is_nullable = False


@NumpyUInt32.nullable
class PandasUInt32(UInt32, backend="pandas"):
    """Pandas 32-bit unsigned integer type."""

    aliases = {pd.UInt32Dtype}
    dtype = pd.UInt32Dtype()
    is_nullable = True


class UInt16(Unsigned):
    """Abstract 16-bit unsigned integer type."""

    aliases = {"uint16", "unsigned int16", "u2"}
    max = 2**16 - 1
    min = 0


@UInt16.default
class NumpyUInt16(UInt16, backend="numpy"):
    """Numpy 16-bit unsigned integer type."""

    aliases = {np.uint16, np.dtype(np.uint16)}
    scalar = np.uint16
    dtype = np.dtype(np.uint16)
    is_nullable = False


@NumpyUInt16.nullable
class PandasUInt16(UInt16, backend="pandas"):
    """Pandas 16-bit unsigned integer type."""

    aliases = {pd.UInt16Dtype}
    dtype = pd.UInt16Dtype()
    is_nullable = True


class UInt8(Unsigned):
    """Abstract 8-bit unsigned integer type."""

    aliases = {"uint8", "unsigned int8", "u1"}
    max = 2**8 - 1
    min = 0


@UInt8.default
class NumpyUInt8(UInt8, backend="numpy"):
    """Numpy 8-bit unsigned integer type."""

    aliases = {np.uint8, np.dtype(np.uint8)}
    dtype = np.dtype(np.uint8)
    is_nullable = False


@NumpyUInt8.nullable
class PandasUInt8(UInt8, backend="pandas"):
    """Pandas 8-bit unsigned integer type."""

    aliases = {pd.UInt8Dtype}
    dtype = pd.UInt8Dtype()
    is_nullable = True


#######################
####    PRIVATE    ####
#######################


# NOTE: some aliases are platform-specific and may be assigned to different
# integer types based on hardware configuration.  Luckily, numpy's dtype()
# factory automatically resolves these, so we can just piggyback off it.


platform_specific_aliases: dict[str, str] = {
    # C char
    "char": str(np.dtype(np.byte)),
    "signed char": "char",
    "byte": "char",
    "b": "char",

    # C short
    "short": str(np.dtype(np.short)),
    "short int": "short",
    "short integer": "short",
    "signed short": "short",
    "signed short int": "short",
    "signed short integer": "short",
    "h": "short",

    # C int
    "intc": str(np.dtype(np.intc)),
    "signed intc": "intc",

    # C long
    "long": str(np.dtype(np.int_)),
    "long int": "long",
    "long integer": "long",
    "signed long": "long",
    "signed long int": "long",
    "signed long integer": "long",
    "l": "long",

    # C long long
    "long long": str(np.dtype(np.longlong)),
    "long long int": "long long",
    "long long integer": "long long",
    "signed long long": "long long",
    "signed long long int": "long long",
    "signed long long integer": "long long",
    "longlong": "long long",
    "signed longlong": "long long",
    "q": "long long",

    # C ssize_t
    "ssize_t": str(np.dtype(np.intp)),
    "intp": "ssize_t",
    "int0": "ssize_t",
    "p": "ssize_t",

    # C unsigned char
    "unsigned char": str(np.dtype(np.ubyte)),
    "unsigned byte": "unsigned char",
    "ubyte": "unsigned char",
    "B": "unsigned char",

    # C unsigned short
    "unsigned short": str(np.dtype(np.ushort)),
    "unsigned short int": "unsigned short",
    "unsigned short integer": "unsigned short",
    "ushort": "unsigned short",
    "H": "unsigned short",

    # C unsigned int
    "unsigned intc": str(np.dtype(np.uintc)),
    "uintc": "unsigned intc",
    "I": "unsigned intc",

    # C unsigned long
    "unsigned long": str(np.dtype(np.uint)),
    "unsigned long int": "unsigned long",
    "unsigned long integer": "unsigned long",
    "ulong": "unsigned long",
    "L": "unsigned long",

    # C unsigned long long
    "unsigned long long": str(np.dtype(np.ulonglong)),
    "unsigned long long int": "unsigned long long",
    "unsigned long long integer": "unsigned long long",
    "ulonglong": "unsigned long long",
    "unsigned longlong": "unsigned long long",
    "Q": "unsigned long long",

    # C size_t
    "size_t": str(np.dtype(np.uintp)),
    "uintp": "size_t",
    "uint0": "size_t",
    "P": "size_t",
}


for alias, lookup in platform_specific_aliases.items():
    REGISTRY.strings[lookup].aliases.add(alias)


REGISTRY.refresh_regex()



"""This module contains dispatched cast() implementations for string data."""
# pylint: disable=unused-argument
import datetime
import re  # normal python regex for compatibility with pd.Series.str.extract
from functools import partial

import dateutil
import numpy as np
import pandas as pd

from pdcast import types
from pdcast.resolve import resolve_type
from pdcast.util.round import Tolerance
from pdcast.util.string import boolean_match
from pdcast.util import time
from pdcast.util.vector import apply_with_errors

from .base import (
    cast, generic_to_boolean, generic_to_complex
)


# TODO: string -> datetime has an extraneous upcast() step.


@cast.overload("string", "bool")
def string_to_boolean(
    series: pd.Series,
    dtype: types.ScalarType,
    true: set,
    false: set,
    errors: str,
    ignore_case: bool,
    **unused
) -> pd.Series:
    """Convert string data to a boolean data type."""
    lookup = dict.fromkeys(true, 1) | dict.fromkeys(false, 0)
    if "*" in true:
        fill = 1  # KeyErrors become truthy
    elif "*" in false:
        fill = 0  # KeyErrors become falsy
    else:
        fill = -1  # raise

    call = partial(
        boolean_match,
        lookup=lookup,
        ignore_case=ignore_case,
        fill=fill
    )
    series = apply_with_errors(series, call, errors=errors)
    series = series.astype(np.bool_)
    return generic_to_boolean(series, dtype, errors=errors)


@cast.overload("string", "int")
def string_to_integer(
    series: pd.Series,
    dtype: types.ScalarType,
    base: int,
    errors: str,
    **unused
) -> pd.Series:
    """Convert string data to an integer data type with the given base."""
    # 2 step conversion: string -> int[python], int[python] -> int
    series = apply_with_errors(series, partial(int, base=base), errors=errors)
    series = series.astype(resolve_type(int).dtype)
    return cast(
        series,
        dtype,
        base=base,
        errors=errors,
        **unused
    )


@cast.overload("string", "float")
def string_to_float(
    series: pd.Series,
    dtype: types.ScalarType,
    tol: Tolerance,
    errors: str,
    **unused
) -> pd.Series:
    """Convert string data to a floating point data type."""
    # 2 step conversion: string -> decimal, decimal -> float
    series = cast(series, "decimal", errors=errors)
    return cast(series, dtype, tol=tol, errors=errors, **unused)


@cast.overload("string", "complex")
def string_to_complex(
    series: pd.Series,
    dtype: types.ScalarType,
    tol: Tolerance,
    downcast: types.CompositeType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert string data to a complex data type."""
    # NOTE: this is technically a 3-step conversion: (1) str -> str (split
    # real/imag), (2) str -> float (x2), (3) float -> complex.  This allows for
    # full precision loss/overflow/downcast checks for both real and imag.

    # (1) separate real, imaginary components via regex
    components = series.str.extract(complex_pattern)
    real_part = components["real"]
    imag_part = components["imag"]

    # (2) convert real, imag to float, applying checks independently
    real_part = cast(
        real_part,
        dtype.equiv_float,
        tol=Tolerance(tol.real),
        downcast=None,
        errors="raise"
    )
    imag_part = cast(
        imag_part,
        dtype.equiv_float,
        tol=Tolerance(tol.imag),
        downcast=None,
        errors="raise"
    )

    # (3) combine floats into complex result
    series = (real_part + imag_part * 1j).astype(dtype.dtype, copy=False)
    return generic_to_complex(
        series,
        dtype,
        tol=tol,
        downcast=downcast,
        errors=errors
    )


@cast.overload("string", "datetime")
def string_to_datetime(
    series: pd.Series,
    dtype: types.ScalarType,
    errors: str,
    **unused
) -> pd.Series:
    """Convert string data into a datetime data type."""
    # NOTE: because this type has no associated scalars, it will never be given
    # as the result of a detect_type() operation.  It can only be specified
    # manually, as the target of a resolve_type() call.

    last_err = None
    for candidate in dtype.larger:
        try:
            return cast(series, candidate, errors="raise", **unused)
        except OverflowError as err:
            last_err = err

    # every candidate overflows - pick the last one and coerce
    if errors == "coerce":
        return cast(
            series,
            dtype,
            errors=errors,
            **unused
        )
    raise last_err


@cast.overload("string", "datetime[pandas]")
def string_to_pandas_timestamp(
    series: pd.Series,
    dtype: types.ScalarType,
    tz: datetime.tzinfo,
    format: str,
    day_first: bool,
    year_first: bool,
    errors: str,
    **unused
) -> pd.Series:
    """Convert datetime strings into pandas Timestamps."""
    # reconcile `tz` argument with timezone attached to dtype, if given
    if tz:
        dtype = dtype.replace(tz=tz)

    # configure kwargs for pd.to_datetime
    kwargs = {
        "dayfirst": day_first,
        "yearfirst": year_first,
        "utc": dtype.tz is not None and time.is_utc(dtype.tz),
        "errors": "raise" if errors == "ignore" else errors
    }
    if format:
        kwargs |= {"format": format, "exact": False}
    else:
        kwargs |= {"format": "mixed"}

    # NOTE: pd.to_datetime() can throw several kinds of errors, some of which
    # are ambiguous.  For simplicity, we catch and re-raise these only as
    # ValueErrors or OverflowErrors.
    try:
        result = pd.to_datetime(series, **kwargs)

    # exception 1: outside pd.Timestamp range, but within datetime.datetime
    except pd._libs.tslibs.np_datetime.OutOfBoundsDatetime as err:
        raise OverflowError(str(err)) from None  # truncate stack

    # exception 2: bad string or outside datetime.datetime range
    except dateutil.parser.ParserError as err:
        raise time.filter_dateutil_parser_error(err) from None

    # account for missing values introduced during coercion
    if errors == "coerce":
        isna = result.isna()
        result = result[~isna]

    # localize to final timezone
    try:
        # NOTE: if utc=False and there are mixed timezones and/or mixed
        # aware/naive strings in the input series, the output of
        # pd.to_datetime() could be malformed.
        if not kwargs["utc"]:
            # homogenous - either naive or consistent timezone
            if pd.api.types.is_datetime64_ns_dtype(result):
                if not result.dt.tz:  # naive
                    result = result.dt.tz_localize(dtype.tz)
                else:  # aware
                    result = result.dt.tz_convert(dtype.tz)

            # non-homogenous - either mixed timezone or mixed aware/naive
            else:  # NOTE: pd.to_datetime() sacrifices ns precision here
                localize = partial(
                    time.localize_pydatetime_scalar,
                    tz=dtype.tz
                )
                # NOTE: np.frompyfunc() implicitly casts to pd.Timestamp
                result = np.frompyfunc(localize, 1, 1)(result)

    # exception 3: overflow induced by timezone localization
    except pd._libs.tslibs.np_datetime.OutOfBoundsDatetime as err:
        raise OverflowError(str(err)) from None

    return result


@cast.overload("string", "datetime[python]")
def string_to_python_datetime(
    series: pd.Series,
    dtype: types.VectorType,
    tz: datetime.tzinfo,
    day_first: bool,
    year_first: bool,
    format: str,
    errors: str,
    **unused
) -> pd.Series:
    """Convert strings into datetime objects."""
    # reconcile `tz` argument with timezone attached to dtype, if given
    if tz:
        dtype = dtype.replace(tz=tz)

    # set up dateutil parserinfo
    parser_info = dateutil.parser.parserinfo(
        dayfirst=day_first,
        yearfirst=year_first
    )

    # apply elementwise
    call = partial(
        time.string_to_pydatetime,
        format=format,
        parser_info=parser_info,
        tz=dtype.tz,
        errors=errors
    )
    series = apply_with_errors(series, call, errors=errors)
    return series.astype(dtype.dtype)


@cast.overload("string", "datetime[numpy]")
def string_to_numpy_datetime64(
    series: pd.Series,
    dtype: types.VectorType,
    format: str,
    tz: datetime.tzinfo,
    errors: str,
    **unused
) -> pd.Series:
    """Convert ISO 8601 strings to a numpy datetime64 data type."""
    if format and not time.is_iso_8601_format_string(format):
        raise TypeError(
            "np.datetime64 strings must be in ISO 8601 format"
        )
    if tz and not time.is_utc(tz):
        raise TypeError(
            "np.datetime64 objects do not carry timezone information"
        )

    # 2-step conversion: string -> ns, ns -> datetime64
    series = apply_with_errors(series, time.iso_8601_to_ns, errors=errors)
    series = series.astype(resolve_type(int).dtype)
    return cast(
        series,
        dtype,
        format=format,
        tz=tz,
        errors=errors,
        **unused
    )


@cast.overload("string", "timedelta")
def string_to_timedelta(
    series: pd.Series,
    dtype: types.ScalarType,
    unit: str,
    step_size: int,
    since: time.Epoch,
    as_hours: bool,
    errors: str,
    **unused
) -> pd.Series:
    """Convert string data into a timedelta representation."""
    # 2-step conversion: str -> int, int -> timedelta
    call = partial(time.timedelta_string_to_ns, as_hours=as_hours, since=since)
    series = apply_with_errors(series, call, errors=errors)
    series = series.astype(resolve_type(int).dtype)
    return cast(
        series,
        dtype,
        unit="ns",
        step_size=1,
        since=since,
        errors=errors,
        **unused
    )


#######################
####    PRIVATE    ####
#######################


complex_pattern = re.compile(
    r"\(?(?P<real>[+-]?[0-9.]+)(?P<imag>[+-][0-9.]+)?j?\)?"
)



"""This module describes a single function (``round_decimal()``), which
performs customizable, vectorized rounding on arbitrary-precision ``Decimal``
values.
"""
import decimal

cimport cython
import numpy as np
cimport numpy as np
import pandas as pd


######################
####    PUBLIC    ####
######################


cpdef object round_decimal(
    object val,
    short decimals,
    str rule
):
    """Performs fast, elementwise rounding on a decimal scalars and arrays,
    according to the given rounding rule.

    Parameters
    ----------
    val : decimal.Decimal | array_like
        The value to round.  If array-like, `val` must have `dtype='O'` and
        contain only `decimal.Decimal` objects.
    rule : str
        A string specifying the rounding strategy to use.  Must be one of
        ('floor', 'ceiling', 'down', 'up', 'half_floor', 'half_ceiling',
        'half_down', 'half_up', 'half_even'), where `up`/`down` round
        away/toward zero, and `ceiling`/`floor` round toward +/- infinity,
        respectively.  Defaults to 'half_even'.
    decimals : int
        The number of decimals to round to.  Positive numbers count to the
        right of the decimal point, and negative values count to the left.
        0 represents rounding in the ones place of `val`.  This follows the
        convention set out in `numpy.around`.  Defaults to 0.

    Returns
    -------
    decimal.Decimal | array_like:
        The result of rounding `val` according to the given rule.

    Raises
    ------
    ValueError:
        If `rule` is not one of the accepted rounding rules ('floor',
        'ceiling', 'down', 'up', 'half_floor', 'half_ceiling', 'half_down',
        'half_up', 'half_even').
    """
    # array
    if isinstance(val, np.ndarray):
        return _round_decimal_array(
            val,
            rule=rule,
            decimals=decimals
        )

    cdef object index

    # series
    if isinstance(val, pd.Series):
        index = val.index
        val = _round_decimal_array(
            val.to_numpy(dtype="O"),
            rule=rule,
            decimals=decimals
        )
        return pd.Series(val, index=index)

    cdef object quantize
    cdef object scale_factor

    # scalar
    if val in decimal_inf_nan:
        return val
    quantize = rounding_rules[rule]
    scale_factor = decimal.Decimal(10) ** decimals
    return quantize(val * scale_factor) / scale_factor



#######################
####    PRIVATE    ####
#######################


cdef set decimal_inf_nan = {
    decimal.Decimal("-inf"),
    decimal.Decimal("inf"),
    decimal.Decimal("nan")
}


cdef dict rounding_rules = {
    "floor": lambda x: x.quantize(1, decimal.ROUND_FLOOR),
    "ceiling": lambda x: x.quantize(1, decimal.ROUND_CEILING),
    "down": lambda x: x.quantize(1, decimal.ROUND_DOWN),
    "up": lambda x: x.quantize(1, decimal.ROUND_UP),
    "half_floor": lambda x: (
        x.quantize(1, decimal.ROUND_HALF_UP) if x < 0 else
        x.quantize(1, decimal.ROUND_HALF_DOWN)
    ),
    "half_ceiling": lambda x: (
        x.quantize(1, decimal.ROUND_HALF_DOWN) if x < 0 else
        x.quantize(1, decimal.ROUND_HALF_UP)
    ),
    "half_down": lambda x: x.quantize(1, decimal.ROUND_HALF_DOWN),
    "half_up": lambda x: x.quantize(1, decimal.ROUND_HALF_UP),
    "half_even": lambda x: x.quantize(1, decimal.ROUND_HALF_EVEN)
}


@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.ndarray[object] _round_decimal_array(
    object[:] arr,
    str rule = "half_even",
    int decimals = 0
):
    """Call the appropriate .quantize() method for each element of input."""
    cdef int arr_length = arr.shape[0]
    cdef object scale_factor = decimal.Decimal(10)**decimals
    cdef int i
    cdef object element
    cdef np.ndarray[object] result = np.empty(arr_length, dtype="O")
    
    quantize = rounding_rules[rule]

    for i in range(arr_length):
        element = arr[i]
        if element in decimal_inf_nan:
            result[i] = element
        else:
            result[i] = quantize(element * scale_factor) / scale_factor
    return result



"""Mypy stubs for pdcast/types/array/object.pyx"""
from typing import Any, Iterable, Iterator, NoReturn

import numpy as np
import pandas as pd
from pandas.core.arrays import ExtensionArray
from pandas.core.dtypes.base import ExtensionDtype

from pdcast.types import ScalarType
from pdcast.util.type_hints import dtype_like


boolean_array = np.ndarray[np.bool_, np.dtype[np.bool_]]


def construct_object_dtype(
    pdcast_type: ScalarType,
    is_boolean: bool,
    is_numeric: bool,
    kind: str = ...,
    nullable: bool = ...,
    common_dtype: dtype_like | None = ...,
) -> ObjectDtype:
    ...
def construct_array_type(pdcast_type: ScalarType) -> type[ExtensionArray]: ...


class ObjectDtype(ExtensionDtype):
    def __init__(self) -> None: ...
    @classmethod
    def construct_from_string(cls, string: str) -> NoReturn: ...
    def __getattr__(self, name: str) -> Any: ...
    def __dir__(self) -> list[str]: ...
    def __repr__(self) -> str: ...


class _ObjectDtype(ObjectDtype):
    _pdcast_type: ScalarType
    name: str
    type: type
    kind: str
    na_value: Any
    _is_boolean: bool
    _is_numeric: bool
    _can_hold_na: bool

    @classmethod
    def construct_array_type(cls) -> type[ExtensionArray]: ...
    def _get_common_dtype(self, dtypes: Iterable[ExtensionDtype]) -> dtype_like | None: ...


class ObjectArray(ExtensionArray):
    __array_priority__: int

    def __init__(
        self, values: Iterable[Any], dtype: dtype_like | None = ..., copy: bool = ...
    ) -> None:
        ...

    ############################
    ####    CONSTRUCTORS    ####
    ############################

    @classmethod
    def _from_sequence(
        cls, scalars: Iterable[Any], dtype: dtype_like | None = ..., copy: bool = ...
    ) -> ObjectArray:
        ...
    @classmethod
    def _from_factorized(
        cls, values: np.ndarray, original: ExtensionArray
    ) -> ObjectArray:
        ...
    @classmethod
    def _concat_same_type(
        cls, to_concat: Iterable[ExtensionArray]
    ) -> ObjectArray:
        ...

    ##########################
    ####    ATTRIBUTES    ####
    ##########################

    @property
    def dtype(self) -> ObjectDtype: ...
    @property
    def _dtype(self) -> ObjectDtype: ...
    @property
    def nbytes(self) -> int: ...

    ########################
    ####    INTERNAL    ####
    ########################

    def _values_for_argsort(self) -> np.ndarray: ...
    def _values_for_factorize(self) -> tuple[np.ndarray, ObjectArray]: ...
    def _reduce(self, name: str, skipna: bool = ..., **kwargs: Any) -> Any: ...
    def __array_ufunc__(
        self, ufunc: np.ufunc, method: str, *inputs: Any, **kwargs: Any
    ) -> ObjectArray:
        ...
    def __arrow_array__(self, type: dtype_like | None = ...) -> Any: ...
    def __from_arrow__(self, array: Any) -> ObjectArray: ...

    ######################
    ####    PUBLIC    ####
    ######################

    def astype(self, dtype: dtype_like, copy: bool = ...) -> ObjectArray: ...
    def isna(self) -> boolean_array: ...
    def take(
        self,
        indices: Iterable[int],
        allow_fill: bool = ...,
        fill_value: Any | None = ...
    ) -> ObjectArray:
        ...
    def copy(self) -> ObjectArray: ...

    ############################
    ####    NEW METHODS    ####
    ############################

    def round(self, *args: Any, **kwargs: Any) -> ObjectArray: ...
    def value_counts(self, dropna: bool = ...) -> pd.Series: ...

    ###############################
    ####    UNARY OPERATORS    ####
    ###############################

    def __abs__(self) -> ObjectArray: ...
    def __neg__(self) -> ObjectArray: ...
    def __pos__(self) -> ObjectArray: ...
    def __invert__(self) -> ObjectArray: ...

    ################################
    ####    BINARY OPERATORS    ####
    ################################

    def __and__(self, other: Any) -> ObjectArray: ...
    def __or__(self, other: Any) -> ObjectArray: ...
    def __xor__(self, other: Any) -> ObjectArray: ...
    def __rshift__(self, other: Any) -> ObjectArray: ...
    def __lshift__(self, other: Any) -> ObjectArray: ...

    ########################################
    ####    REVERSE BINARY OPERATORS    ####
    ########################################

    def __rand__(self, other: Any) -> ObjectArray: ...
    def __ror__(self, other: Any) -> ObjectArray: ...
    def __rxor__(self, other: Any) -> ObjectArray: ...
    def __rrshift__(self, other: Any) -> ObjectArray: ...
    def __rlshift__(self, other: Any) -> ObjectArray: ...

    #########################################
    ####    IN-PLACE BINARY OPERATORS    ####
    #########################################

    def __iand__(self, other: Any) -> ObjectArray: ...
    def __ior__(self, other: Any) -> ObjectArray: ...
    def __ixor__(self, other: Any) -> ObjectArray: ...
    def __irshift__(self, other: Any) -> ObjectArray: ...
    def __ilshift__(self, other: Any) -> ObjectArray: ...

    ##############################
    ####    MATH OPERATORS    ####
    ##############################

    def __add__(self, other: Any) -> ObjectArray: ...
    def __sub__(self, other: Any) -> ObjectArray: ...
    def __mul__(self, other: Any) -> ObjectArray: ...
    def __matmul__(self, other: Any) -> ObjectArray: ...
    def __truediv__(self, other: Any) -> ObjectArray: ...
    def __floordiv__(self, other: Any) -> ObjectArray: ...
    def __mod__(self, other: Any) -> ObjectArray: ...
    def __divmod__(self, other: Any) -> ObjectArray: ...
    def __pow__(self, other: Any, mod: Any | None = ...) -> ObjectArray: ...

    ######################################
    ####    REVERSE MATH OPERATORS    ####
    ######################################

    def __radd__(self, other: Any) -> ObjectArray: ...
    def __rsub__(self, other: Any) -> ObjectArray: ...
    def __rmul__(self, other: Any) -> ObjectArray: ...
    def __rmatmul__(self, other: Any) -> ObjectArray: ...
    def __rtruediv__(self, other: Any) -> ObjectArray: ...
    def __rfloordiv__(self, other: Any) -> ObjectArray: ...
    def __rmod__(self, other: Any) -> ObjectArray: ...
    def __rdivmod__(self, other: Any) -> ObjectArray: ...
    def __rpow__(self, other: Any, mod: Any | None = ...) -> ObjectArray: ...

    #######################################
    ####    IN-PLACE MATH OPERATORS    ####
    #######################################

    def __iadd__(self, other: Any) -> ObjectArray: ...
    def __isub__(self, other: Any) -> ObjectArray: ...
    def __imul__(self, other: Any) -> ObjectArray: ...
    def __imatmul__(self, other: Any) -> ObjectArray: ...
    def __itruediv__(self, other: Any) -> ObjectArray: ...
    def __ifloordiv__(self, other: Any) -> ObjectArray: ...
    def __imod__(self, other: Any) -> ObjectArray: ...
    def __ipow__(self, other: Any, mod: Any | None = ...) -> ObjectArray: ...

    ####################################
    ####    COMPARISON OPERATORS    ####
    ####################################

    def __lt__(self, other: Any) -> boolean_array: ...
    def __le__(self, other: Any) -> boolean_array: ...
    def __eq__(self, other: Any) -> boolean_array: ...
    def __ne__(self, other: Any) -> boolean_array: ...
    def __gt__(self, other: Any) -> boolean_array: ...
    def __ge__(self, other: Any) -> boolean_array: ...

    #################################
    ####    CONTAINER METHODS    ####
    #################################

    def __getitem__(self, key: Any) -> Any | ObjectArray: ...
    def __setitem__(self, key: Any, value: Any) -> None: ...
    def __delitem__(self, key: Any) -> NoReturn: ...
    def __iter__(self) -> Iterator[Any]: ...
    def __len__(self) -> int: ...
    def __contains__(self, item: Any) -> bool: ...



"""This package contains vectorized rounding operations for numeric data,
with customizable rounding rules and tolerances.

Modules
-------
decimal
    Customizable rounding for ``Decimal`` objects.

float
    Customizable rounding for floating point numbers.

integer
    Vectorized integer division with customizable rounding.

tolerance
    A ``Tolerance`` object for ``snap()`` operations.

Constants
---------
valid_rules
    A tuple listing the various rounding rules that are accepted by this
    package.
"""
from .decimal import round_decimal
from .float import round_float
from .integer import round_div
from .tolerance import Tolerance



# pylint: disable=redefined-outer-name, unused-argument
from __future__ import annotations
import datetime
from functools import partial

from pdcast.decorators.attachable import attachable
from pdcast.decorators.dispatch import dispatch
from pdcast.decorators.extension import extension_func
from pdcast.detect import detect_type
from pdcast.util import time
from pdcast.util.vector import apply_with_errors


@attachable
@extension_func
@dispatch("series")
def tz_convert(
    series: pd.Series,
    tz: str | datetime.tzinfo | None,
    **unused
) -> pd.Series:
    """TODO"""
    raise NotImplementedError(
        f"{detect_type(series)} objects do not carry timezone information"
    )


#########################
####    ARGUMENTS    ####
#########################


tz_convert.argument(time.tz)


#######################
####    PRIVATE    ####
#######################


@tz_convert.overload("datetime[pandas]")
def localize_pandas_timestamp(
    series: pd.Series,
    tz: datetime.tzinfo | None,
    **unused
) -> pd.Series:
    """TODO"""
    series = series.astype(detect_type(series).dtype, copy=False)
    original = getattr(series.dt.tz_convert, "original", series.dt.tz_convert)
    return original(tz, **unused)


@tz_convert.overload("datetime[python]")
def localize_python_datetime(
    series: pd.Series,
    tz: datetime.tzinfo | None,
    **unused
) -> pd.Series:
    """TODO"""
    # emulate pandas tz_convert limitation
    if not detect_type(series).tz:
        raise TypeError(
            "Cannot convert tz-naive Timestamp, use tz_localize to localize"
        )

    series_type = detect_type(series)
    call = partial(time.localize_pydatetime_scalar, tz=tz)
    series = apply_with_errors(series, call, errors="raise")
    return series.astype(series_type.replace(tz=tz).dtype)



"""This module contains all the prepackaged string types for the ``pdcast``
type system.
"""
from sys import getsizeof

import numpy as np
import pandas as pd

from .base import DTYPE, Type, TypeMeta


class String(Type):
    """Abstract string type."""

    aliases = {
        "string", "str", "unicode", "str0", "str_", "unicode_", "U", pd.StringDtype
    }

    @classmethod
    def from_dtype(cls, dtype: DTYPE) -> TypeMeta:
        """TODO"""
        if isinstance(dtype, pd.StringDtype):
            if dtype.storage == "pyarrow":
                return PyArrowString

            return PythonString

        return cls


@String.default
class PythonString(String, backend="python"):
    """Python string type."""

    aliases = {str, np.str_, np.dtype("U")}
    scalar = str
    dtype = pd.StringDtype("python")
    itemsize = getsizeof("")
    missing = pd.StringDtype("python").na_value


class PyArrowString(String, backend="pyarrow"):
    """PyArrow string type."""

    aliases = set()
    dtype = pd.StringDtype("pyarrow")
    itemsize = getsizeof("")
    missing = pd.StringDtype("pyarrow").na_value



"""This module contains all the prepackaged decimal types for the ``pdcast``
type system.
"""
import decimal

from .base import Type


# https://github.com/pandas-dev/pandas/blob/e246c3b05924ac1fe083565a765ce847fcad3d91/pandas/tests/extension/decimal/array.py


class Decimal(Type):
    """Abstract decimal type."""

    aliases = {"decimal"}


@Decimal.default
class PythonDecimal(Decimal, backend="python"):
    """Python decimal type."""

    aliases = {decimal.Decimal}
    scalar = decimal.Decimal
    missing = decimal.Decimal("nan")



"""This module describes a single function (``round_div()``), which mimics the
integer division operator `//` for vectorized data, with customizable rounding
behavior.
"""
import numpy as np
import pandas as pd


######################
####    PUBLIC    ####
######################


cpdef object round_div(
    object numerator,
    object denominator,
    str rule = "floor"
):
    """Vectorized integer division with customizable rounding behavior.

    Unlike other approaches, this function does not perform float conversion
    at any point.  Instead, it replicates each rounding rule by adding a
    simple integer bias at each index before applying the `//` floor division
    operator.  This allows it to retain full integer precision for arbitrary
    choices of `numerator` and `denominator`.

    Parameters
    ----------
    numerator : int | array_like
        Integer numerator.  Can be vectorized (`np.ndarray`/`pd.Series`),
        with arbitrary dimensions.
    denominator : int | array_like
        Integer denominator.  Can be vectorized (`np.ndarray`/`pd.Series`),
        with arbitrary dimensions.
    rule : str (default 'floor')
        A string specifying the rounding strategy to use.  Must be one of
        ('floor', 'ceiling', 'down', 'up', 'half_floor', 'half_ceiling',
        'half_down', 'half_up', 'half_even'), where `down`/`up` round
        toward/away from zero, and `floor`/`ceiling` round toward -/+ infinity,
        respectively.  Defaults to 'floor', which matches the behavior of the
        base `//` operator.

    Returns
    -------
    int | np.ndarray:
        The result of integer division with the specified rounding rule.  If
        either of the numeric inputs are vectorized, the result will match the
        input type.

    Raises
    ------
    ValueError:
        If `rule` is not one of the accepted rounding rules ('floor',
        'ceiling', 'down', 'up', 'half_floor', 'half_ceiling', 'half_down',
        'half_up', 'half_even').
    """
    # NOTE: // does not automatically broadcast pd.Series objects
    if (
        isinstance(numerator, pd.Series) and
        isinstance(denominator, pd.Series) and
        numerator.shape != denominator.shape
    ):
        raise ValueError(
            f"operands could not be broadcast together with shapes "
            f"{numerator.shape} {denominator.shape}"
        )

    if rule == "floor":  # no bias to add, just use // directly
        return numerator // denominator

    cdef object bias
    cdef tuple valid_rules
    cdef str err_msg

    try:  # get numerator bias for given rounding rule
        bias = integer_rounding_bias[rule](numerator, denominator)
    except KeyError as err:
        valid_rules = ('floor',) + tuple(integer_rounding_bias)
        err_msg = f"`rule` must be one of {valid_rules}, not {repr(rule)}"
        raise ValueError(err_msg) from err

    return (numerator + bias) // denominator


#######################
####    PRIVATE    ####
#######################


cdef object _bias_down(object numer, object denom):
    """Apply floor where `numer` and `d` have the same sign, and ceiling where
    they differ.  This is equivalent to:

    `((numer < 0) ^ (denom < 0)) * (denom - sign(denom))`
    """
    cdef object neg_d = (denom < 0)
    cdef object bias = denom + neg_d - (denom > 0)
    bias *= (numer < 0) ^ neg_d
    return bias


cdef object _bias_up(object numer, object denom):
    """Apply ceiling where `numer` and `denom` have the same sign, and floor
    where they differ.  This is equivalent to:

    `((numer > 0) ^ (denom < 0)) * (denom - sign(denom))`
    """
    cdef object neg_d = (denom < 0)
    cdef object bias = denom + neg_d - (denom > 0)
    bias *= (numer > 0) ^ neg_d
    return bias


cdef object _bias_half_down(object numer, object denom):
    """Apply half_floor where `numer` and `denom` have the same sign, and
    half_ceiling where they differ.  This is equivalent to:

    `(denom + (denom < 0) - ((numer > 0) ^ (denom < 0)) * sign(denom)) // 2`
    """
    cdef object neg_d = (denom < 0)
    cdef object bias = denom + neg_d
    bias -= ((numer > 0) ^ neg_d) * (-1 * neg_d + (denom > 0))
    bias //= 2
    return bias


cdef object _bias_half_up(object numer, object denom):
    """Apply half_ceiling where `numer` and `denom` have the same sign, and
    half_floor where they differ.  This is equivalent to:

    `(denom + (denom < 0) - ((numer < 0) ^ (denom < 0)) * sign(denom)) // 2`
    """
    cdef object neg_d = (denom < 0)
    cdef object bias = denom + neg_d
    bias -= ((numer < 0) ^ neg_d) * (-1 * neg_d + (denom > 0))
    bias //= 2
    return bias


cdef object _bias_half_even(object numer, object denom):
    """Apply half_ceiling where the quotient `numer // denom` would be odd, and
    half_floor where it would be even.  This is equivalent to:

    `(denom + (denom < 0) + ((numer // denom) % 2 - 1) * sign(denom)) // 2`
    """
    cdef object neg_d = (denom < 0)
    cdef object bias = denom + neg_d
    bias += ((numer // denom) % 2 - 1) * (-1 * neg_d + (denom > 0))
    bias //= 2
    return bias


cdef dict integer_rounding_bias = {
    "ceiling":      lambda n, d: d - (d > 0) + (d < 0),
    "down":         _bias_down,
    "up":           _bias_up,
    "half_floor":   lambda n, d: (d - (d > 0)) // 2 + (d < 0),
    "half_ceiling": lambda n, d: (d + (d < 0)) // 2,
    "half_down":    _bias_half_down,
    "half_up":      _bias_half_up,
    "half_even":    _bias_half_even
}



"""This module contains all the prepackaged timedelta types for the ``pdcast``
type system.
"""
import datetime
import re

import numpy as np
import pandas as pd

# from bertrand.util import time

from .base import Type, TypeMeta


# TODO:
# >>> Timedelta["numpy", "5ns"].params
# mappingproxy({'unit': None, 'step_size': None})
# -> should be: mappingproxy({'unit': 'ns', 'step_size': 5})


class Timedelta(Type):
    """Abstract timedelta type."""

    aliases = {"timedelta"}


@Timedelta.default
class PandasTimedelta(Timedelta, backend="pandas"):
    """Pandas timedelta type."""

    aliases = {pd.Timedelta, "pandas.Timedelta", "pd.Timedelta"}
    scalar = pd.Timedelta
    dtype = np.dtype("m8[ns]")
    itemsize = 8
    max = pd.Timedelta.max.value
    min = pd.Timedelta.min.value
    missing = pd.NaT



class PythonTimedelta(Timedelta, backend="python"):
    """Python timedelta type."""

    aliases = {datetime.timedelta, "pytimedelta", "datetime.timedelta"}
    scalar = datetime.timedelta
    # max = time.pytimedelta_to_ns(datetime.timedelta.max)
    # min = time.pytimedelta_to_ns(datetime.timedelta.min)
    missing = pd.NaT


class NumpyTimedelta64(Timedelta, backend="numpy", cache_size=64):
    """Numpy timedelta64 type."""

    # NOTE: dtype is set to object due to pandas and its penchant for
    # automatically converting datetimes to pd.Timestamp.  Otherwise, we'd use
    # an ObjectDtype or the raw numpy dtypes here.

    aliases = {
        np.timedelta64, np.dtype("m8"), "m8", "timedelta64", "numpy.timedelta64",
        "np.timedelta64",
    }
    scalar = np.timedelta64
    dtype = np.dtype(object)  # workaround for above
    itemsize = 8
    na_value = np.timedelta64("NaT")

    def __class_getitem__(
        cls,
        unit: str | None = None,
        step_size: int | None = None
    ) -> TypeMeta:
        if unit is None:
            # inverted max, min always triggers upcast mechanism during conversions
            return cls.flyweight(unit, None, max=0, min=1)

        # parse numpy-style m8 string
        m8 = m8_pattern.match(unit)
        if m8 is None:
            raise ValueError(f"invalid unit: '{unit}'")

        p_unit = m8.group("unit")
        if step_size is None:
            p_step = int(m8.group("step_size") or 1)
        elif m8.group("step_size"):
            raise ValueError(
                f"conflicting units: '{unit}' != '{p_unit}, {step_size}'"
            )
        else:
            if step_size is None:
                p_step = 1
            elif step_size < 1:
                raise ValueError(f"step size must be positive: {step_size}")
            else:
                p_step = step_size

        # # NOTE: these epochs are chosen to minimize range in the event of irregular
        # # units ('Y'/'M'), so that conversions work regardless of leap days and
        # # irregular month lengths
        # _max = time.convert_unit(
        #     2**63 - 1,
        #     p_unit,
        #     "ns",
        #     since=time.Epoch(np.datetime64("2001-02-01"))
        # )
        # _min = time.convert_unit(
        #     -2**63 + 1,  # NOTE: -2**63 reserved for NaT
        #     p_unit,
        #     "ns",
        #     since=time.Epoch(np.datetime64("2000-02-01"))
        # )

        # return cls.flyweight(unit, step_size, max=_max, min=_min)
        return cls.flyweight(p_unit, p_step)

    @classmethod
    def from_scalar(cls, scalar: np.timedelta64) -> TypeMeta:
        """Parse a scalar m8 value according to unit, step size."""
        unit, step_size = np.datetime_data(scalar)
        return cls.flyweight(unit, step_size)

    @classmethod
    def from_dtype(cls, dtype: np.dtype[np.timedelta64]) -> TypeMeta:
        """Translate a numpy m8 dtype into the pdcast type system."""
        unit, step_size = np.datetime_data(dtype)

        # # NOTE: pandas uses numpy m8 dtypes for its own Timedelta type, so we
        # # need to check the type of the array to detect this case.
        # if isinstance(array, (pd.Index, pd.Series)):
        #     # TODO: include non-ns units?
        #     return PandasTimedeltaType

        return cls.flyweight(None if unit == "generic" else unit, step_size)

    @classmethod
    def from_string(
        cls,
        unit: str | None = None,
        step_size: str | None = None
    ) -> TypeMeta:
        """Parse an m8 string in the type specification mini-language.

        Numpy timedeltas support two different parametrized syntaxes:
        
            1.  Numpy format, which concatenates step size and unit into a
                single field (e.g. 'm8[5ns]').
            2.  pdcast format, which lists each field individually (e.g.
                'timedelta[numpy, ns, 5]').  This matches the output of the
                str() function for these types.
        """
        if step_size is None:
            return cls[unit, step_size]

        return cls[unit, int(step_size)]

    # @property
    # def larger(self) -> Iterator:
    #     """If no original unit is given, iterate through each one in order."""
    #     if self.unit is None:
    #         yield from (self(unit=unit) for unit in time.valid_units)
    #     else:
    #         yield from ()


#######################
####    PRIVATE    ####
#######################


# REGISTRY.edges.add(PythonTimedelta, NumpyTimedelta64)


# regex to parse numpy-style "m8[{step_size}{unit}]" strings
m8_pattern: re.Pattern[str] = re.compile(
    r"(?P<step_size>[0-9]+)?(?P<unit>ns|us|ms|s|m|h|D|W|M|Y)"
)



"""This module describes an ``@extension_func`` decorator that transforms an
ordinary Python function into one that can accept managed arguments with custom
validators and dynamic values.
"""
from __future__ import annotations
from functools import wraps
import inspect
import threading
from types import MappingProxyType
from typing import Any, Callable, Mapping, get_type_hints

from .base import EMPTY, Arguments, FunctionDecorator, Signature


######################
####    PUBLIC    ####
######################


def extension_func(func: Callable) -> Callable:
    """A decorator that allows a function to accept managed arguments with
    dynamic values.

    Parameters
    ----------
    func : Callable
        A Python function or other callable to be decorated.  If this accepts a
        ``**kwargs`` dict or similar variable-length keyword argument, then it
        will be allowed to take on
        :ref:`extension arguments <extension_func.extension>` that are defined
        at runtime.

    Returns
    -------
    ExtensionFunc
        A :ref:`cooperative <attachable.nested>` decorator that allows
        transparent access to the decorated function.  These objects manage
        default values and argument validators for the decorated callable.

    Notes
    -----
    The returned :class:`ExtensionFunc <pdcast.ExtensionFunc>` inherits from
    :class:`threading.local <python:threading.local>`, which isolates its
    default values to the current thread.  Each instance inherits the default
    values of the main thread at the time it is constructed.
    """
    main_thread = []  # using a one-element list bypasses UnboundLocalError

    class _ExtensionFunc(ExtensionFunc):
        """A subclass of :class:`ExtensionFunc <pdcast.ExtensionFunc>` that
        supports dynamic assignment of ``@properties`` without affecting other
        instances.
        """

        def __init__(self, _func: Callable):
            super().__init__(_func)

            # store attributes from main thread
            if threading.current_thread() == threading.main_thread():
                main_thread.append({
                    "class": type(self),
                    "signature": self._signature,
                })

            # load attributes from main thread
            else:
                main = main_thread[0]
                self._signature.copy_settings(main["signature"])
                for k in main["signature"].validators:
                    setattr(type(self), k, getattr(main["class"], k))

    return _ExtensionFunc(func)


#######################
####    PRIVATE    ####
#######################


class ExtensionFunc(FunctionDecorator, threading.local):
    """A wrapper for a function that manages its arguments.

    Parameters
    ----------
    func : Callable
        The decorated function or other callable.

    Notes
    -----
    Whenever an argument is :meth:`registered <ExtensionFunc.argument>` with
    this function, it is added as a managed :class:`property <python:property>`
    with appropriate getter, setter, and deleter methods.  These are
    automatically derived from the validation function itself, and can be used
    to manage its default value externally, without touching any hard code.

    Additionally, this object inherits from :class:`threading.local`, which
    means that each instance is isolated to the current thread.  If the
    function is referenced from a child thread, a new instance will be created
    with the arguments and settings of the main thread.  This instance can then
    be modified without affecting the behavior of any other threads.

    Examples
    --------
    See the :ref:`API docs <extension_func>` for example usage.
    """

    _reserved = FunctionDecorator._reserved | {"_signature"}

    def __init__(self, func: Callable):
        super().__init__(func=func)
        self._signature = ExtensionSignature(func)

    ####################
    ####    BASE    ####
    ####################

    # NOTE: @properties will be added to this class by the @argument decorator

    @property
    def arguments(self) -> MappingProxyType:
        """A mapping of all managed arguments to their respective validators.

        Returns
        -------
        MappingProxyType
            A read-only dictionary mapping argument names to their associated
            validation functions.

        Examples
        --------
        .. doctest::

            >>> @extension_func
            ... def foo(bar, baz, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> @foo.argument
            ... def bar(val, context: dict) -> int:
            ...     return int(val)

            >>> foo.arguments   # doctest: +SKIP
            mappingproxy({'bar': <function bar at 0x7ff5ad9c6e60>})
        """
        return MappingProxyType(self._signature.validators)

    @property
    def settings(self) -> Mapping[str, Any]:
        """A mapping of all managed arguments to their current values.

        Returns
        -------
        MappingProxyType
            A read-only dictionary mapping argument names to their current
            values for this :class:`ExtensionFunc <pdcast.ExtensionFunc>`.  If
            no explicit value is associated with an argument, then it will be
            excluded from the map.

        Examples
        --------
        .. doctest::

            >>> @extension_func
            ... def foo(bar, baz=2, **kwargs)
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> @foo.argument(default=1)
            ... def bar(val, context: dict) -> int:
            ...     return int(val)

            >>> @foo.argument
            ... def baz(val, context: dict) -> int:
            ...     return int(val)

            >>> foo.settings
            mappingproxy({'bar': 1, 'baz': 2})
            >>> foo.bar, foo.baz = 24, "-17"
            >>> foo.settings
            mappingproxy({'bar': 24, 'baz': -17})
        """
        return self._signature.settings

    def argument(
        self,
        func: Callable | None = None,
        *,
        name: str | None = None,
        default: Any = EMPTY
    ) -> Callable:
        """A decorator that transforms a validation function into a managed
        argument for this :class:`ExtensionFunc <pdcast.ExtensionFunc>`.

        Parameters
        ----------
        name : str | None, default None
            The name of the argument that the validator validates.  If this is
            left as :data:`None <python:None>`, then the name of the validator
            will be used instead.
        default : Any | None, default EMPTY
            The default value to use for this argument.  This is implicitly
            passed to the validator itself, so any custom logic that is
            implemented there will also be applied to this value if it is not
            empty.  Otherwise, if this argument is omitted and the decorated
            function defines a default value in its call signature, then that
            value will be used instead.

        Returns
        -------
        Callable
            A decorated version of the validation function that automatically
            fills out its second positional argument (typically a ``context``
            dict) with a table containing the current value of each argument
            at the time the function is invoked.  This validator will
            implicitly be called whenever the
            :class:`ExtensionFunc <pdcast.ExtensionFunc>` is executed with the
            named argument as input.

        Raises
        ------
        TypeError
            If the decorated object is not callable or its ``name`` conflicts
            with a built-in attribute of the
            :class:`ExtensionFunc <pdcast.ExtensionFunc>` or one of its
            derivatives.  This can also be raised if the function does not
            implement a ``**kwargs`` argument and ``name`` is absent from its
            argument list.
        KeyError
            If a managed property of the same name already exists.

        Notes
        -----
        Validation function must accept at least two positional arguments as
        shown:

        .. code:: python

            def validator(val, context: dict):
                ...

        Where ``val`` is an arbitrary input to the argument and ``context`` is
        a dictionary containing the current values for each argument at the
        time the :class:`ExtensionFunc <pdcast.ExtensionFunc>` was invoked.  If
        the validator interacts with other arguments (to enforce mutual
        exclusivity, for example), then they can be obtained from the
        ``context`` dictionary.

        .. note::

            Race conditions may be introduced when arguments access each other
            in their validators.  This can be mitigated somewhat by using safe
            access methods like :meth:`dict.get() <python:dict.get>` rather
            than direct indexing, as well as by avoiding modifications to the
            ``context`` dictionary at runtime.

            Additionally, the values that are obtained from the ``context``
            dictionary are not guaranteed to have been validated prior to
            accessing them.  If you need to ensure that a value is valid before
            using it, then you can do so by manually applying the same
            coercions as in the referenced argument's validation function.

        Examples
        --------
        See the :ref:`API docs <extension_func.arguments>` for example usage.
        """

        def decorator(validator: Callable) -> Callable:
            """Attach a validation function to the ExtensionFunc as a managed
            property.
            """
            # ensure validator is valid
            self._signature.check_validator(validator)

            # use name of validator as argument name if not explicitly given
            _name = name
            if _name is None:
                _name = validator.__name__
            elif not isinstance(_name, str):
                raise TypeError(f"name must be a string, not {type(_name)}")

            # ensure name is not already in use
            if _name in self._signature.validators:
                raise KeyError(f"argument '{_name}' already exists")

            # ensure name is not overwriting a reserved attribute
            if _name in dir(self):
                raise TypeError(f"'{_name}' is a reserved attribute")

            # wrap validator to accept argument context
            @wraps(validator)
            def validate_context(val, context=self.settings, **kwargs):
                """Automatically populate `context` argument of validator."""
                return validator(val, context, **kwargs)

            # add argument to signature and generate @property
            prop = self._signature.register_argument(
                name=_name,
                validator=validate_context,
                default=default,
                annotation=get_type_hints(validator)["return"]
            )
            setattr(type(self), _name, prop)

            # return wrapped validator
            return validate_context

        if func is None:
            return decorator
        return decorator(func)

    def remove_arg(self, name: str) -> None:
        """Remove a managed argument from an
        :class:`ExtensionFunc <pdcast.ExtensionFunc>`.

        Parameters
        ----------
        name : str
            The name of the argument to remove.  This must be contained in the
            :attr:`arguments <pdcast.ExtensionFunc.arguments>` attribute of
            the :class:`ExtensionFunc <pdcast.ExtensionFunc>`.

        Raises
        ------
        KeyError
            If the argument is not being actively managed by this
            :class:`ExtensionFunc <pdcast.ExtensionFunc>`.

        Examples
        --------
        .. doctest::

            >>> @extension_func
            ... def foo(bar, baz, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> @foo.argument(default=1)
            ... def bar(val, context: dict) -> int:
            ...     return int(val)

            >>> foo.bar
            1
            >>> foo.remove_arg("bar")
            >>> foo.bar
            Traceback (most recent call last):
                ...
            AttributeError: 'function' object has no attribute 'bar'.
        """
        # ensure argument is being actively managed
        if name not in self._signature.validators:
            raise KeyError(f"'{name}' is not a managed argument")

        # commit to deletion
        delattr(type(self), name)
        self._signature.defaults.pop(name, None)
        self._signature.validators.pop(name)

        # replace with original parameter if present
        try:
            self._signature.reset_parameter(name)
        except TypeError:
            self._signature.delete_parameter(name)

    def reset_defaults(self) -> None:
        """Reset all arguments to their original defaults.

        Notes
        -----
        This is effectively equivalent to calling the ``del`` keyword on every
        argument registered to this
        :class:`ExtensionFunc <pdcast.ExtensionFunc>`.

        Examples
        --------
        .. doctest::

            >>> @extension_func
            ... def foo(bar, baz, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> @foo.argument(default=1)
            ... def bar(val, context: dict) -> int:
            ...     return int(val)

            >>> @foo.argument(default=2)
            ... def baz(val, context: dict) -> int:
            ...     return int(val)

            >>> foo.bar, foo.baz
            (1, 2)
            >>> foo.bar, foo.baz = 18, -34
            >>> foo.bar, foo.baz
            (12, -34)
            >>> foo.reset_defaults()
            >>> foo.bar, foo.baz
            (1, 2)
        """
        self._signature.reset_defaults()

    ###############################
    ####    SPECIAL METHODS    ####
    ###############################

    def __call__(self, *args, **kwargs) -> Any:
        """Execute the decorated function with the managed arguments.

        Parameters
        ----------
        *args, **kwargs
            Arbitrary positional and keyword arguments passed to this function.
            These will always be redirected through their respective validation
            function before being passed to the decorated callable.

        Returns
        -------
        Any
            The return value of the decorated callable with the validated
            arguments.

        Notes
        -----
        Validators are only called for the arguments that are explicitly passed
        to this method.  If an argument is not passed, then its default value
        will be used instead, and the validator will not be invoked.

        Examples
        --------
        See the :ref:`API docs <extension_func.arguments>` for example usage.
        """
        # generate ExtensionArguments
        bound = self._signature(*args, **kwargs)

        # pass arguments through their respective validators
        bound.validate()

        # NOTE: validate() implicitly applies defaults

        # invoke wrapped function
        return self.__wrapped__(*bound.args, **bound.kwargs)

    def __repr__(self) -> str:
        """Reconstruct the function's effective signature using the current
        value of each argument.

        Returns
        -------
        str
            A string describing the state of the function's arguments,
            including their current settings and any
            :ref:`extensions <extension_func.extension>` that have been added.

        Examples
        --------
        .. doctest::

            >>> @extension_func
            ... def foo(bar, baz, **kwargs):
            ...     return {"bar": bar, "baz": baz} | kwargs

            >>> @foo.argument(default=1)
            ... def bar(val, context: dict) -> int:
            ...     return int(val)

            >>> @foo.argument(default=2)
            ... def baz(val, context: dict) -> int:
            ...     return int(val)

            >>> @foo.argument(default=3)
            ... def qux(val, context: dict) -> int:
            ...     return int(val)

            >>> foo
            foo(bar=1, baz=2, *, qux=3, **kwargs)
            >>> repr(foo)
            'foo(bar=1, baz=2, *, qux=3, **kwargs)'
        """
        return self._signature.reconstruct(annotations=False)


class ExtensionSignature(Signature):
    """A wrapper around an :class:`inspect.Signature <python:inspect.Signature>`
    object that serves as a factory for
    :class:`ExtensionArguments <pdcast.ExtensionArguments>`.

    Parameters
    ----------
    func : Callable
        The function whose signature will be wrapped.

    Notes
    -----
    These capture the arguments that are passed to
    :class:`ExtensionFunc.__call__() <pdcast.ExtensionFunc.__call__>` and
    encapsulate them in a dedicated
    :class:`ExtensionArguments <pdcast.ExtensionArguments>` object.

    :class:`ExtensionArguments <pdcast.ExtensionArguments>` provide a
    high-level interface for interacting with and validating arguments, without
    having to interact with the complicated machinery of the
    :mod:`inspect <python:inspect>` module.
    """

    def __init__(self, func: Callable):
        super().__init__(func=func)
        self.accepts_kwargs = any(
            par.kind == par.VAR_KEYWORD for par in self.parameters
        )
        self.defaults: dict[str, Any] = {
            par.name: par.default for par in self.parameters
            if par.default is not EMPTY
        }
        self.validators: dict[str, Callable] = {}

    @property
    def settings(self) -> Mapping[str, Any]:
        """A mapping of all managed arguments to their current values.

        Returns
        -------
        MappingProxyType
            A read-only dictionary mapping argument names to their current
            values for this
            :class:`ExtensionSignature <pdcast.ExtensionSignature>`.  If no
            explicit value is associated with an argument, then it will be
            excluded from the map.
        """
        return MappingProxyType({
            par.name: par.default for par in self.parameters
            if par.name in self.validators and par.default is not EMPTY
        })

    def copy_settings(self, other: ExtensionSignature) -> None:
        """Copy the current settings of another
        :class:`ExtensionSignature <pdcast.ExtensionSignature>`.

        Parameters
        ----------
        ExtensionSignature
            The signature whose settings will be copied.

        Notes
        -----
        This is used to copy settings from the main thread's
        :class:`ExtensionFunc <pdcast.ExtensionFunc>` to its children when a
        new thread is spawned.
        """
        self.sig = other.sig  # copy other signature
        self.validators = other.validators.copy()
        self.defaults = other.defaults.copy()

    def reset_defaults(self) -> None:
        """Reset all settings to their hardcoded defaults.

        Notes
        -----
        This resets to either the ``default`` that was given in
        :meth:`@argument <pdcast.ExtensionFunc.argument>` or the hardcoded
        value in the function's original signature, whichever comes first.
        """
        self.parameters = tuple(
            par.replace(default=self.defaults.get(par.name, EMPTY))
            for par in self.parameters
        )

    def check_validator(self, validator: Callable) -> None:
        """Confirm that a validator is callable with the expected parameters.

        Parameters
        ----------
        validator : Callable
            A custom validation function decorated with
            :meth:`@argument <pdcast.ExtensionFunc.argument>`.

        Raises
        ------
        TypeError
            If the validator is not callable, or if it does not accept at least
            2 arguments: the value itself and a context dictionary storing the
            values of every other argument that the function was invoked with.
        """
        # ensure validator is callable
        if not callable(validator):
            raise TypeError(f"validator must be callable: {validator}")

        # ensure validator accepts at least 2 arguments
        if len(inspect.signature(validator).parameters) < 2:
            raise TypeError(
                f"validator must accept at least 2 arguments: {validator}"
            )

    def register_argument(
        self,
        name: str,
        validator: Callable,
        default: Any = EMPTY,
        annotation: Any = EMPTY
    ) -> property:
        """Add an argument to this signature.

        Parameters
        ----------
        name : str
            The name of the argument to add.
        validator : Callable
            The validation function to use when validating and setting the
            argument.  This must be decorated with
            :meth:`@ExtensionFunc.argument <pdcast.ExtensionFunc.argument>`.

        Returns
        -------
        property
            A managed :class:`property <python:property>` to be added to the
            :class:`ExtensionFunc <pdcast.ExtensionFunc>` itself.  This has
            full getter, setter, and deleter logic based on the ``validator``
            that was passed as input.

        Raises
        ------
        TypeError
            If the argument name does not appear in the function's signature
            and the function does not accept a variable-length ``**kwargs``
            argument (or a differently-named equivalent).

        Notes
        -----
        The returned :class:`property <python:property>` uses the following
        logic for attribute access:

            *   When :meth:`__get__() <python:object.__get__>` is called, the
                result is always equal to the default value of the associated
                :class:`Parameter <python:inspect.Parameter>` for this
                :class:`ExtensionSignature <pdcast.ExtensionSignature>`.
            *   When :meth:`__set__() <python:object.__set__>` is called, the
                assigned value will be passed through the validator and then
                stored directly on the argument's
                :class:`Parameter <python:inspect.Parameter>`.
            *   When :meth:`__del__() <python:object.__del__>` is called, the
                argument will be reset to its original state, removing any
                dynamic values that were assigned to it.
        """
        # pass default value through validator
        if default is EMPTY:
            pars = self.parameter_map
            if name in pars and pars[name].default is not EMPTY:
                default = validator(pars[name].default)
        else:
            default = validator(default)

        # if argument is in signature, modify its default
        if name in self.parameter_map:
            self.set_parameter(name, default=default, annotation=annotation)

        # otherwise, extend signature with new argument
        else:
            # assert signature accepts a `**kwargs`-equivalent
            if not self.accepts_kwargs:
                raise TypeError(
                    f"'{self.__qualname__}()' has no argument '{name}'"
                )

            # generate parameter and insert into signature
            self.add_parameter(
                name,
                kind="KEYWORD_ONLY",
                default=default,
                annotation=annotation
            )

        # remember validator, default
        self.validators[name] = validator
        if default is not EMPTY:
            self.defaults[name] = default

        # generate getter, setter, and deleter for @property
        def getter(self) -> Any:
            """Get the value of a managed argument."""
            result = self._signature.parameter_map[name].default
            if result is EMPTY:
                raise TypeError(f"'{name}' has no default value")
            return result

        def setter(self, val: Any) -> None:
            """Set the value of a managed argument."""
            self._signature.set_parameter(name, default=validator(val))

        def deleter(self) -> None:
            """Replace the value of a managed argument with its default."""
            val = self._signature.defaults.get(name, EMPTY)
            if val is not EMPTY:
                setter(self, val)
            else:
                self._signature.set_parameter(name, default=EMPTY)

        return property(getter, setter, deleter, doc=validator.__doc__)

    # pylint: disable=no-self-argument
    def __call__(__self, *args, **kwargs) -> ExtensionArguments:
        """Bind the arguments to this signature and return a corresponding
        :class:`ExtensionArguments <pdcast.ExtensionArguments>` object.

        Parameters
        ----------
        *args, **kwargs
            Arbitrary positional and keyword arguments to be bound to this
            signature.

        Returns
        -------
        ExtensionArguments
            A new :class:`ExtensionArguments <pdcast.ExtensionArguments>`
            object that encapsulates the bound arguments.

        Notes
        -----
        This is always called on the input to
        :meth:`ExtensionFunc.__call__() <pdcast.ExtensionFunc.__call__>`.
        """
        bound = __self.sig.bind_partial(*args, **kwargs)
        return ExtensionArguments(bound=bound, signature=__self)


class ExtensionArguments(Arguments):
    """Captured arguments from a call to
    :meth:`ExtensionFunc(*args, **kwargs) <pdcast.ExtensionFunc.__call__>`.

    Notes
    -----
    Whenever an :class:`ExtensionFunc <pdcast.ExtensionFunc>` is called, its
    arguments are intercepted and encapsulated in an
    :class:`ExtensionArguments <pdcast.ExtensionArguments>` object.  These
    behave similarly to
    :class:`inspect.BoundArguments <python:inspect.BoundArguments>`, but with
    additional operations for validating their values and applying dynamic
    defaults.
    """

    def validate(self) -> None:
        """Pass each argument through its associated validator.

        Notes
        -----
        This should always be called before
        :meth:`apply_defaults() <pdcast.Arguments.apply_defaults>` in order to
        minimize validation calls.  Since defaults are pre-validated by their
        associated properties, there is no need to check them manually.
        """
        # record original arguments before applying defaults
        explicit = tuple(self.values)

        # apply defaults.  These are pre-validated by their properties.
        self.apply_defaults()

        # validate the non-default arguments
        for name in explicit:
            if name in self.signature.validators:
                validator = self.signature.validators[name]
                self.values[name] = validator(self.values[name], self.values)



from .calendar cimport (
    days_per_year, days_per_4_years, days_per_100_years, days_per_400_years,
    days_per_month, days_in_month, date_to_days, days_to_date, is_leap_year,
    leaps_between
)
from .datetime cimport (
    filter_dateutil_parser_error, is_iso_8601_format_string,
    iso_8601_format_pattern, iso_8601_pattern, iso_8601_to_ns, is_utc,
    ns_to_pydatetime, numpy_datetime64_to_ns, pandas_timestamp_to_ns,
    parser_overflow_pattern, py_aware_utc, py_naive_utc, pydatetime_to_ns,
    string_to_pydatetime
)
from .epoch cimport Epoch, epoch_aliases
from .timedelta cimport (
    numpy_timedelta64_to_ns, pandas_timedelta_to_ns, pytimedelta_to_ns,
    timedelta_regex
)
from .unit cimport (
    as_ns, convert_unit, round_months_to_ns, round_years_to_ns, valid_units
)



"""This module describes a ``CategoricalType`` object, which can be used to
dynamically wrap other types.
"""
from __future__ import annotations

from typing import Any, Iterable

import pandas as pd

from .base import EMPTY, Decorator, DecoratorMeta, Empty, TypeMeta, UnionMeta, detect
from .base.meta import TOKEN


# TODO: Categorical should be able to accept Unions?
# NOTE: this is enabled in pandas, but maybe shouldn't be here.
# -> If CategoricalType becomes a Composite Pattern decorator, then this
# might be handled automatically.


class Categorical(Decorator, cache_size=256):
    """Categorical decorator for :class:`ScalarType` objects.

    This decorator keeps track of categorical levels for series objects of the
    wrapped type.
    """

    aliases = {pd.CategoricalDtype, "Category", "category", "categorical"}

    def __class_getitem__(
        cls,
        wrapped: TypeMeta | DecoratorMeta | None = None,
        levels: Iterable[Any] | Empty = EMPTY
    ) -> DecoratorMeta:
        if wrapped is None:
            raise NotImplementedError("TODO")  # TODO: should never occur

        if levels is EMPTY:
            slug_repr = levels
            dtype = pd.CategoricalDtype()  # NOTE: auto-detects levels when used
            patch = wrapped
            while not patch.is_default:
                patch = patch.as_default
            dtype._bertrand_wrapped_type = patch  # disambiguates wrapped type
        else:
            levels = wrapped(levels)
            slug_repr = list(levels)
            dtype = pd.CategoricalDtype(levels)

        return cls.flyweight(wrapped, slug_repr, dtype=dtype, levels=levels)

    @classmethod
    def from_string(cls, wrapped: str, levels: str | Empty = EMPTY) -> DecoratorMeta:
        """Resolve a categorical specifier in the type specification
        mini-language.
        """
        if levels is EMPTY:
            return cls[wrapped, levels]

        if levels.startswith("[") and levels.endswith("]"):
            stripped = levels[1:-1].strip()
        elif levels.startswith("(") and levels.endswith(")"):
            stripped = levels[1:-1].strip()
        else:
            raise TypeError(f"invalid levels: {repr(levels)}")

        return cls[wrapped, list(s.group() for s in TOKEN.finditer(stripped))]

    @classmethod
    def from_dtype(cls, dtype: DTYPE) -> DecoratorMeta | UnionMeta:
        """Convert a pandas CategoricalDtype into a
        :class:`CategoricalType <pdcast.CategoricalType>` object.
        """
        if dtype.categories is None:
            if hasattr(dtype, "_bertrand_wrapped_type"):
                return dtype._bertrand_wrapped_type
            return cls

        wrapped = detect(dtype.categories)
        if isinstance(wrapped, UnionMeta):
            # TODO: produce a correct index for the original array.
            # -> this can maybe be a special case in detect()
            raise NotImplementedError(
                "Mixed-type categoricals are not currently supported"
            )

        return cls[wrapped, dtype.categories]

    @classmethod
    def transform(cls, series: pd.Series[Any]) -> pd.Series[Any]:
        """Convert a series into a categorical representation."""
        # TODO: this should use a @dispatch function to allow for custom behavior
        # -> Ideally, this would just use a generalized overload of cast()

        if cls.dtype.categories is None:
            dtype = pd.CategoricalDtype(cls.wrapped(series.unique()))
        else:
            dtype = cls.dtype
        return series.astype(dtype, copy=False)
